@@ -33,27 +33,23 @@ only rendering itself: the preparation of the target surface must be done
 elsewhere).
 """
 
-from __future__ import print_function, unicode_literals
-
 import logging
 logger = logging.getLogger(__name__)
 
-import os
 import math
 import enum
+import pathlib
 import tempfile
 import mimetypes
 import webbrowser
+import collections
+from urllib.request import url2pathname
+from urllib.parse import urlsplit
 
 import gi
 gi.require_version('Poppler', '0.18')
 from gi.repository import Poppler
 
-try:
-    from urllib.request import url2pathname
-except ImportError:
-    from urllib import url2pathname
-
 from pympress.util import fileopen
 
 
@@ -233,6 +229,10 @@ class Link(object):
         return lambda *a, **k: fun(*(tuple(args) + tuple(a)), **dict(kwargs, **k))
 
 
+#: A class that holds all the properties for media files
+Media = collections.namedtuple('Media', ['relative_margins', 'filename', 'autoplay', 'repeat', 'poster',
+                                         'show_controls'], defaults=[False, False, False, False])
+
 
 class Page(object):
     """ Class representing a single page.
@@ -254,7 +254,7 @@ class Page(object):
     page_label = None
     #: All the links in the page, as a `list` of :class:`~pympress.document.Link` instances
     links = []
-    #: All the media in the page, as a `list` of tuples of (area, filename)
+    #: All the media in the page, as a `list` of :class:`~pympress.document.Media`
     medias = []
     #: `float`, page width
     pw = 0.
@@ -290,30 +290,35 @@ class Page(object):
 
         # Read annotations, in particular those that indicate media
         for annotation in self.page.get_annot_mapping():
-            content = annotation.annot.get_contents()
-            if content:
-                self.annotations.append(content)
-
             annot_type = annotation.annot.get_annot_type()
+
             if annot_type == Poppler.AnnotType.LINK:
                 # just an Annot, not subclassed -- probably redundant with links
                 continue
             elif annot_type == Poppler.AnnotType.MOVIE:
                 movie = annotation.annot.get_movie()
                 filepath = self.parent.get_full_path(movie.get_filename())
-                if filepath:
-                    # TODO there is no autoplay, or repeatCount
-                    relative_margins = Poppler.Rectangle()
-                    relative_margins.x1 = annotation.area.x1 / self.pw        # left
-                    relative_margins.x2 = 1.0 - annotation.area.x2 / self.pw  # right
-                    relative_margins.y1 = annotation.area.y1 / self.ph        # bottom
-                    relative_margins.y2 = 1.0 - annotation.area.y2 / self.ph  # top
-                    media = (relative_margins, filepath, movie.show_controls())
-                    self.medias.append(media)
-                    action = Link.build_closure(self.parent.play_media, hash(media))
-                else:
+                if not filepath:
                     logger.error(_("Pympress can not find file ") + movie.get_filename())
                     continue
+
+                relative_margins = Poppler.Rectangle()
+                relative_margins.x1 = annotation.area.x1 / self.pw        # left
+                relative_margins.x2 = 1.0 - annotation.area.x2 / self.pw  # right
+                relative_margins.y1 = annotation.area.y1 / self.ph        # bottom
+                relative_margins.y2 = 1.0 - annotation.area.y2 / self.ph  # top
+
+                movie_options = {'show_controls': movie.show_controls(), 'poster': movie.need_poster()}
+                try:
+                    movie_options['repeat'] = movie.get_play_mode() == Poppler.MoviePlayMode.REPEAT
+                    # NB: autoplay not part of Poppler’s MovieActivationParameters struct
+                except AttributeError:
+                    pass  # Missing functions in pre-21.04 Poppler versions
+
+                media = Media(relative_margins, filepath, **movie_options)
+                self.medias.append(media)
+                action = Link.build_closure(self.parent.play_media, hash(media))
+
             elif annot_type == Poppler.AnnotType.SCREEN:
                 action_obj = annotation.annot.get_action()
                 if not action_obj:
@@ -323,18 +328,22 @@ class Page(object):
                     continue
             elif annot_type == Poppler.AnnotType.FILE_ATTACHMENT:
                 attachment = annotation.annot.get_attachment()
-                prefix, ext = os.path.splitext(attachment.name)
-                with tempfile.NamedTemporaryFile('wb', suffix=ext, prefix=prefix, delete=False) as f:
+                filename = pathlib.Path(attachment.name)
+                with tempfile.NamedTemporaryFile('wb', suffix=filename.suffix, prefix=filename.stem, delete=False) as f:
                     # now the file name is shotgunned
-                    filename = f.name
+                    filename = pathlib.Path(f.name)
                     self.parent.remove_on_exit(filename)
-                if not attachment.save(filename):
+                if not attachment.save(str(filename)):
                     logger.error(_("Pympress can not extract attached file"))
                     continue
                 action = Link.build_closure(fileopen, filename)
             elif annot_type in {Poppler.AnnotType.TEXT, Poppler.AnnotType.POPUP,
                                 Poppler.AnnotType.FREE_TEXT}:
-                # text-only annotations, hide them from screen
+                # text-only annotations, hide them from screen and show them in annotations popup
+                content = annotation.annot.get_contents()
+                if content:
+                    self.annotations.append(annotation.annot)
+
                 self.page.remove_annot(annotation.annot)
                 continue
             elif annot_type in {Poppler.AnnotType.STRIKE_OUT, Poppler.AnnotType.HIGHLIGHT,
@@ -462,9 +471,9 @@ class Page(object):
                 ext = get_extension(media.get_mime_type())
                 with tempfile.NamedTemporaryFile('wb', suffix=ext, prefix='pdf_embed_', delete=False) as f:
                     # now the file name is shotgunned
-                    filename = f.name
+                    filename = pathlib.Path(f.name)
                     self.parent.remove_on_exit(filename)
-                if not media.save(filename):
+                if not media.save(str(filename)):
                     logger.error(_("Pympress can not extract embedded media"))
                     return None
             else:
@@ -473,14 +482,21 @@ class Page(object):
                     logger.error(_("Pympress can not find file ") + media.get_filename())
                     return None
 
-            # TODO grab the show_controls, autoplay, repeat
             relative_margins = Poppler.Rectangle()
             relative_margins.x1 = rect.x1 / self.pw        # left
             relative_margins.x2 = 1.0 - rect.x2 / self.pw  # right
             relative_margins.y1 = rect.y1 / self.ph        # bottom
             relative_margins.y2 = 1.0 - rect.y2 / self.ph  # top
 
-            media = (relative_margins, filename, False)
+            media_options = {}
+            try:
+                media_options['autoplay'] = media.get_auto_play()
+                media_options['show_controls'] = media.get_show_controls()
+                media_options['repeat'] = media.get_repeat_count() - 1
+                # NB: no poster in Poppler’s MediaParameters
+            except AttributeError:
+                pass
+            media = Media(relative_margins, filename, **media_options)
             self.medias.append(media)
             return Link.build_closure(self.parent.play_media, hash(media))
 
@@ -560,6 +576,57 @@ class Page(object):
         return self.annotations
 
 
+    def new_annotation(self, pos, rect=None):
+        """ Add an annotation to this page
+
+        Args:
+            pos (`int`): The position in the list of annotations in which to insert this annotation
+            rect (:class:`~Poppler.Rectangle`): A rectangle for the position of this annotation
+
+        Returns:
+            :class:`~Poppler.Annot`: A new annotation on this page
+        """
+        if pos < 0:
+            pos = 0
+        if pos > len(self.annotations):
+            pos = len(self.annotations)
+
+        if rect is None:
+            rect = Poppler.Rectangle()
+            rect.x1 = self.pw - 20
+            rect.x2 = rect.x1 + 20
+            rect.y2 = self.ph - len(self.annotations) * 20
+            rect.y1 = rect.y2 - 20
+
+        new_annot = Poppler.AnnotText.new(self.parent.doc, rect)
+        new_annot.set_icon(Poppler.ANNOT_TEXT_ICON_NOTE)
+        self.annotations.insert(pos, new_annot)
+        self.parent.made_changes()
+        return new_annot
+
+
+    def set_annotation(self, pos, value):
+        """ Add an annotation to this page
+
+        Args:
+            pos (`int`): The number of the annotation
+            value (`str`): The new contents of the annotation
+        """
+        rect = self.annotations[pos].get_rectangle()
+        self.remove_annotation(pos)
+        self.new_annotation(pos, rect).set_contents(value)
+
+
+    def remove_annotation(self, pos):
+        """ Add an annotation to this page
+
+        Args:
+            pos (`int`): The number of the annotation
+        """
+        self.parent.made_changes()
+        del self.annotations[pos]
+
+
     def get_media(self):
         """ Get the list of medias this page might want to play.
 
@@ -626,7 +693,9 @@ class Document(object):
 
     #: Current PDF document (:class:`~Poppler.Document` instance)
     doc = None
-    #: Path to pdf
+    #: `str` full path to pdf
+    uri = None
+    #: :class:`~pathlib.Path` to pdf if uri is a file: URI
     path = None
     #: Number of pages in the document
     nb_pages = -1
@@ -634,7 +703,7 @@ class Document(object):
     #: navigation in the document faster by avoiding calls to Poppler when loading
     #: a page that has already been loaded.
     pages_cache = {}
-    #: Files that are temporary and need to be removed
+    #: `set` of :class:`~pathlib.Path` representing the temporary files which need to be removed
     temp_files = set()
     #: History of pages we have visited
     history = []
@@ -644,6 +713,8 @@ class Document(object):
     page_labels = []
     #: `bool` indicating whether the second half of pages are in fact notes pages
     notes_after = False
+    #: `bool` indicating whether there were modifications to the document
+    changes = False
 
     #: callback, to be connected to :func:`~pympress.extras.Media.play`
     play_media = lambda *args: None
@@ -664,6 +735,15 @@ class Document(object):
         # Setup PDF file
         self.uri = uri
         self.doc = pop_doc
+        self.changes = False
+
+        if uri is not None:
+            uri_parts = urlsplit(uri, scheme='file')
+            self.path = pathlib.Path(url2pathname(uri_parts.path))
+            if uri_parts.scheme == 'file':
+                self.path = pathlib.Path.cwd().joinpath(self.path.name)
+        else:
+            self.path = None
 
         # Pages number
         if pop_doc is not None:
@@ -700,12 +780,12 @@ class Document(object):
         while True:
             action = index_iter.get_action()
             title = ''
+            page = None
             try:
                 if action.type == Poppler.ActionType.GOTO_DEST:
                     title = action.goto_dest.title
                     if action.goto_dest.dest.type == Poppler.DestType.NAMED:
-                        dest = self.doc.find_dest(action.goto_dest.dest.named_dest)
-                        page = dest.page_num - 1
+                        page = self.doc.find_dest(action.goto_dest.dest.named_dest).page_num - 1
                     elif action.goto_dest.dest.type == Poppler.DestType.UNKNOWN:
                         raise AssertionError('Unknown type of destination')
                     else:
@@ -726,11 +806,11 @@ class Document(object):
 
             # there should not be synonymous sections, correct the page here to a better guess
             if page in index:
-                lower_bound = max(index)
+                lower_bound = max(index.keys())
                 find = index[lower_bound]
                 while 'children' in find:
-                    lower_bound = max(find)
-                    find = find[lower_bound]
+                    lower_bound = max(find['children'].keys())
+                    find = find['children'][lower_bound]
 
                 try:
                     page = min(number for number, label in enumerate(self.page_labels)
@@ -769,6 +849,49 @@ class Document(object):
         return doc
 
 
+    def made_changes(self):
+        """ Notify the document that some changes were made (e.g. annotations edited)
+        """
+        self.changes = True
+
+
+    def has_changes(self):
+        """ Return whether that some changes were made (e.g. annotations edited)
+        """
+        return self.changes
+
+
+    def save_changes(self, dest_uri=None):
+        """ Save the changes
+
+        Args:
+            dest_uri (`str` or `None`): The URI where to save the file, or None to save in-place
+        """
+        if self.doc is None:
+            return
+
+        for page in self.pages_cache.values():
+            for annot in page.get_annotations():
+                page.page.add_annot(annot)
+
+        if dest_uri is not None and dest_uri != self.uri:
+            if self.doc.save(dest_uri):
+                self.changes = False
+        else:
+            # We can’t overwrite the current file directly, so create a temporary file and then overwrite
+            with tempfile.NamedTemporaryFile('wb', suffix=self.path.suffix, prefix=self.path.stem, dir=self.path.parent,
+                                             delete=False) as f:
+                temp_path = pathlib.Path(f.name)
+
+            if self.doc.save(temp_path.as_uri()):
+                self.changes = False
+                temp_path.replace(self.path)
+
+        for page in self.pages_cache.values():
+            for annot in page.get_annotations():
+                page.page.remove_annot(annot)
+
+
     def guess_notes(self, horizontal, vertical, current_page=0):
         """ Get our best guess for the document mode.
 
@@ -1011,18 +1134,18 @@ class Document(object):
         """ Returns full path, extrapolated from a path relative to this document or to the current directory.
 
         Args:
-            filename (`str`):  Name of the file or relative path to it
+            filename (:class:`~pathlib.Path` or `str`):  Name of the file or relative path to it
 
         Returns:
-            `str`: the full path to the file or None if it doesn't exist
+            :class:`~pathlib.Path`: the full path to the file or None if it doesn't exist
         """
-        filepath = None
-        if os.path.isabs(filename):
-            return os.path.normpath(filename) if os.path.exists(filename) else None
+        filename = pathlib.Path(filename)
+        if filename.is_absolute():
+            return filename.resolve() if filename.exists() else None
 
-        for d in [os.path.dirname(url2pathname(self.uri.split('://', 1)[1])), os.getcwd()]:
-            filepath = os.path.normpath(os.path.join(d, filename))
-            if os.path.exists(filepath):
+        for dirname in [pathlib.Path(url2pathname(urlsplit(self.uri).path)).parent, pathlib.Path.cwd()]:
+            filepath = dirname.joinpath(filename)
+            if filepath.exists():
                 return filepath
 
 
@@ -1030,7 +1153,7 @@ class Document(object):
         """ Remember a temporary file to delete later.
 
         Args:
-            filename (`str`): The path to the file to delete
+            filename (:class:`~pathlib.Path`): The path to the file to delete
         """
         self.temp_files.add(filename)
 
@@ -1039,7 +1162,8 @@ class Document(object):
         """ Removes all files that were extracted from the pdf into the filesystem.
         """
         for f in self.temp_files:
-            os.remove(f)
+            if f.exists():
+                f.unlink()
         self.temp_files.clear()
 
 
