@@ -23,207 +23,79 @@
 -----------------------------------------------------------------------------------------------------
 """
 
-from __future__ import print_function, unicode_literals
-
 import logging
 logger = logging.getLogger(__name__)
 
-import os.path
+import pathlib
+import mimetypes
+import functools
+from urllib.request import url2pathname
+
 import gi
 import cairo
 gi.require_version('Gtk', '3.0')
 from gi.repository import Gtk, Gdk, GLib, Gio
 
-import mimetypes
-import functools
-
-try:
-    from urllib.request import url2pathname
-except ImportError:
-    from urllib import url2pathname
-
-from pympress import document, builder
-
-
-class TimingReport(builder.Builder):
-    """ Widget tracking and displaying hierachically how much time was spent in each page/section of the presentation.
-    """
-    #: `list` of time at which each page was reached
-    page_time = []
-    #: `int` the time at which the clock was reset
-    end_time = -1
-    #: The :class:`~Gtk.TreeView` containing the timing data to display in the dialog
-    timing_treeview = None
-    #: A :class:`~Gtk.Dialog` to contain the timing to show
-    time_report_dialog = None
-    #: `bool` marking whether next page transition should reset the history of page timings
-    clear_on_next_transition = False
-
-    #: A `dict` containing the structure of the current document
-    doc_structure = {}
-    #: A `list` with the page label of each page of the current document
-    page_labels = []
-    #: `bool` tracking whether a document is opened
-    document_open = False
-
-    def __init__(self, parent):
-        super(TimingReport, self).__init__()
-        self.load_ui('time_report_dialog')
-        self.time_report_dialog.set_transient_for(parent.p_win)
-        self.time_report_dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
-
-        self.connect_signals(self)
-        parent.setup_actions({
-            'timing-report': dict(activate=self.show_report),
-        })
-
-
-    def transition(self, page, time):
-        """ Record a transition time between slides.
-
-        Args:
-            page (`int`): the page number of the current slide
-            time (`int`): the number of seconds elapsed since the beginning of the presentation
-        """
-        if not self.document_open:
-            return
-
-        if self.clear_on_next_transition:
-            self.clear_on_next_transition = False
-            del self.page_time[:]
-
-        self.page_time.append((page, time))
-
-
-    def reset(self, reset_time):
-        """ A timer reset. Clear the history as soon as we start changing pages again.
-        """
-        self.end_time = reset_time
-        self.clear_on_next_transition = True
-
-
-    @staticmethod
-    def format_time(secs):
-        """ Formats a number of seconds as `minutes:seconds`.
-
-        Returns:
-            `str`: The formatted time, with 2+ digits for minutes and 2 digits for seconds.
-        """
-        return '{:02}:{:02}'.format(*divmod(secs, 60))
-
-
-    def set_document_metadata(self, doc_structure, page_labels):
-        """ Show the popup with the timing infortmation.
-
-        Args:
-            doc_structure (`dict`): the structure of the document
-            page_labels (`list`): the page labels for each of the pages
-        """
-        self.document_open = len(page_labels) != 0
-
-        # Do not update if we only close the document.
-        # That way, the report is still accessible when the document is closed.
-        if not self.document_open:
-            return
-
-        self.doc_structure = doc_structure
-        self.page_labels = page_labels
-
-        # Clear the report when there is a new document opened.
-        del self.page_time[:]
-
-
-    def show_report(self, gaction, param=None):
-        """ Show the popup with the timing infortmation.
-        """
-        times = [time for page, time in self.page_time]
-        durations = (e - s for s, e in zip(times, times[1:] + [self.end_time]))
-
-        min_time = min(time for page, time in self.page_time) if self.page_time else 0
-        infos = {'time': min_time, 'duration': 0, 'children': [], 'page': 0}
-        infos['title'] = 'Full presentation'
-
-        for (page, start_time), duration in zip(self.page_time, durations):
-            if not duration:
-                continue
-
-            infos['duration'] += duration
-
-            # lookup the position of the page in the document structure (section etc)
-            lookup = self.doc_structure
-            cur_info_pos = infos
-            while lookup:
-                try:
-                    pos = max(p for p in lookup if p <= page)
-                except ValueError:
-                    break
-                item = lookup[pos]
-                lookup = item.get('children', None)
-
-                if cur_info_pos['children'] and cur_info_pos['children'][-1]['page'] == pos:
-                    cur_info_pos['children'][-1]['duration'] += duration
-                else:
-                    cur_info_pos['children'].append({'page': pos, 'title': item['title'], 'children': [],
-                                                     'duration': duration, 'time': start_time})
-                cur_info_pos = cur_info_pos['children'][-1]
-
-            # add the actual page as a leaf node
-            label = self.page_labels[page] if 0 <= page < len(self.page_labels) else 'None'
-            cur_info_pos['children'].append({'page': page, 'title': _('slide #') + label,
-                                             'duration': duration, 'time': start_time})
-
-
-        treemodel = self.timing_treeview.get_model()
-        if treemodel:
-            treemodel.clear()
-
-        treemodel = Gtk.TreeStore(str, str, str, str)
-
-        dfs_info = [(None, infos)]
-        while dfs_info:
-            first_it, first = dfs_info.pop()
-            page = first['page']
-            label = self.page_labels[page] if 0 <= page < len(self.page_labels) else 'None'
-
-            last_col = '{} ({}/{})'.format(label, page, len(self.page_labels))
-            row = [first['title'], self.format_time(first['time']), self.format_time(first['duration']), last_col]
-            it = treemodel.append(first_it, row)
-
-            if 'children' in first:
-                dfs_info.extend((it, child) for child in reversed(first['children']))
-
-        self.timing_treeview.set_model(treemodel)
-        self.timing_treeview.expand_row(Gtk.TreePath.new_first(), False)
-
-        self.time_report_dialog.run()
-        self.time_report_dialog.hide()
+from pympress import document, builder, util
 
 
 class Annotations(object):
     """ Widget displaying a PDF’s text annotations.
     """
-    #: The containing :class:`~Gtk.TextView` widget for the annotations
-    annotations_textview = None
-    #: :class:`~Gtk.ScrolledWindow` making the annotations list scroll if it's too long
-    scrolled_window = None
+    #: The containing :class:`~Gtk.TreeView` widget for the annotations
+    annotations_treeview = None
+    #: The containing :class:`~Gtk.ListStore` storing the annotations to be displayed
+    annotations_liststore = None
+    #: The :class:`~Gtk.Entry` in which we are currently editing an annotation, or None
+    editing = None
+
+    new_doc_annotation = lambda *args: None
+    set_doc_annotation = lambda *args: None
+    remove_doc_annotation = lambda *args: None
 
     def __init__(self, builder):
         super(Annotations, self).__init__()
         builder.load_widgets(self)
+        builder.setup_actions({
+            'add-annotation': dict(activate=self.add_annotation),
+            'remove-annotation': dict(activate=self.remove_annotation),
+        })
 
 
-    def add_annotations(self, annotations):
+    def load_annotations(self, annot_page):
         """ Add annotations to be displayed (typically on going to a new slide).
 
         Args:
-            annotations (`list`): A list of strings, that are the annotations to be displayed
+            annot_page (:class:`~pympress.document.Page`): The page object that contains the annotations
         """
-        buf = self.annotations_textview.get_buffer()
-        buf.set_text('\n'.join(annotations))
+        self.annotations_liststore.clear()
+        for num, annot in enumerate(annot_page.get_annotations()):
+            self.annotations_liststore.append([annot.get_contents(), num])
+
+        self.new_doc_annotation = annot_page.new_annotation
+        self.set_doc_annotation = annot_page.set_annotation
+        self.remove_doc_annotation = annot_page.remove_annotation
 
 
-    def on_scroll(self, widget, event):
-        """ Try scrolling the annotations window.
+    def try_cancel(self):
+        """ Try to cancel editing
+
+        Returns:
+            `bool`: whether editing was enabled and thus canceled
+        """
+        if self.editing is None:
+            return False
+
+        rows = self.annotations_treeview.get_selection().get_selected_rows()[1]
+        if rows:
+            self.annotations_treeview.set_cursor(rows[0], None, False)
+        return True
+
+
+    def key_event(self, widget, event):
+        """ Handle a key (press/release) event.
+
+        Needed to forward events directly to the :class:`~Gtk.Entry`, bypassing the global action accelerators.
 
         Args:
             widget (:class:`~Gtk.Widget`):  the widget which has received the event.
@@ -232,16 +104,79 @@ class Annotations(object):
         Returns:
             `bool`: whether the event was consumed
         """
-        adj = self.scrolled_window.get_vadjustment()
-        if event.direction == Gdk.ScrollDirection.UP:
-            adj.set_value(adj.get_value() - adj.get_step_increment())
-        elif event.direction == Gdk.ScrollDirection.DOWN:
-            adj.set_value(adj.get_value() + adj.get_step_increment())
-        else:
+        if self.editing is None:
             return False
+        elif event.get_event_type() == Gdk.EventType.KEY_PRESS:
+            return self.editing.do_key_press_event(self.editing, event)
+        elif event.get_event_type() == Gdk.EventType.KEY_RELEASE:
+            return self.editing.do_key_release_event(self.editing, event)
+        return False
+
+
+    def editing_started(self, cell_renderer, widget, entry_number):
+        """ Handle edit start
+
+        Args:
+            cell_renderer (:class:`~Gtk.CellRenderer`): The renderer which received the signal
+            widget (:class:`~Gtk.CellEditable`): the Gtk entry editing the annotation entry
+            entry_number (`str`): the string representation of the path identifying the edited cell
+        """
+        self.editing = widget
+
+
+    def editing_validated(self, cell_renderer, entry_number, new_content):
+        """ Handle successful edit: store the new cell value in the model and the document
+
+        Args:
+            cell_renderer (:class:`~Gtk.CellRenderer`): The renderer which received the signal
+            entry_number (`str`): the string representation of the path identifying the edited cell
+            new_content (`str`): the new value of the edited cell
+        """
+        path = Gtk.TreePath.new_from_string(entry_number)
+        row = self.annotations_liststore.get_iter(path)
+        self.annotations_liststore.set_value(row, 0, new_content)
+        self.set_doc_annotation(path.get_indices()[0], new_content)
+        self.editing_finished(cell_renderer)
+
+
+    def editing_finished(self, cell_renderer):
+        """ Handle the end of editing
+
+        Args:
+            cell_renderer (:class:`~Gtk.CellRenderer`): The renderer which received the signal
+        """
+        self.editing = None
+
+
+    def add_annotation(self, gaction, param=None):
+        """ Add an annotation to the the page’s annotation list
+
+        Args:
+            gaction (:class:`~Gio.Action`): the action triggering the call, which identifies which backend
+            param (:class:`~GLib.Variant`): an optional parameter
+        """
+        path = self.annotations_liststore.get_path(self.annotations_liststore.append())
+        self.annotations_treeview.set_cursor(path, self.annotations_treeview.get_columns()[0], True)
+        self.new_doc_annotation(path.get_indices()[0])
         return True
 
 
+    def remove_annotation(self, gaction, param=None):
+        """ Remove an annotation to the from the page’s annotation list
+
+        Args:
+            gaction (:class:`~Gio.Action`): the action triggering the call, which identifies which backend
+            param (:class:`~GLib.Variant`): an optional parameter
+        """
+        rows = self.annotations_treeview.get_selection().get_selected_rows()[1]
+        if not rows:
+            return False
+        self.annotations_liststore.remove(self.annotations_liststore.get_iter(rows[0]))
+        self.remove_doc_annotation(rows[0].get_indices()[0])
+        return True
+
+
+
 class Media(object):
     """ Class managing statically the medias and media player backends, to enable play/pause callbacks.
 
@@ -331,16 +266,16 @@ class Media(object):
 
         self.remove_media_overlays()
 
-        for relative_margins, filename, show_controls in current_page.get_media():
-            media_id = hash((relative_margins, filename, show_controls))
+        for media in current_page.get_media():
+            media_id = hash(media)
 
             if media_id not in self._media_overlays:
-                mime_type, enc = mimetypes.guess_type(filename)
+                mime_type, enc = mimetypes.guess_type(media.filename)
                 factory = self.get_factory(mime_type)
 
                 if not factory:
                     logger.warning('No available overlay for mime type {}, ignoring media {}'
-                                   .format(mime_type, filename))
+                                   .format(mime_type, media.filename))
                     continue
 
                 action_group = Gio.SimpleActionGroup.new()
@@ -351,21 +286,17 @@ class Media(object):
                     'set_time': dict(activate=functools.partial(self.set_time, media_id), parameter_type=float)
                 }, action_group)
 
-                v_da_c = factory(self.c_overlay, show_controls, relative_margins, page_type, action_group)
-                v_da_p = factory(self.p_overlay, True, relative_margins, page_type, action_group)
-
-                v_da_c.set_file(filename)
-                v_da_p.set_file(filename)
+                v_da_c = factory(self.c_overlay, page_type, action_group, media)
+                v_da_p = factory(self.p_overlay, page_type, action_group, media._replace(show_controls=True))
 
                 self._media_overlays[media_id] = (v_da_c, v_da_p)
 
             self._media_overlays[media_id][0].mute(True)
             self._media_overlays[media_id][1].mute(False)
 
-            for w in self._media_overlays[media_id]:
-                if w.autoplay:
-                    self.set_time(media_id, param=GLib.Variant.new_double(0))
-                    w.show()
+            if any(overlay.autoplay for overlay in self._media_overlays[media_id]):
+                GLib.idle_add(self.play, media_id)
+            # TODO: handle poster
 
 
     def resize(self, which=None):
@@ -464,33 +395,35 @@ class Media(object):
             logger.info(_('Caused by ') + type(e).__name__ + ': ' + str(e))
 
 
-        try:
-            from pympress.media_overlays.gst_backend import GstOverlay
-
-            gst_version = GstOverlay.setup_backend(self.conf.getlist('gst', 'init_options'))
-            self._backends['gst'] = GstOverlay
-            if self.conf.getboolean('gst', 'enabled'):
-                self.types_list['gst'] = self.conf.getlist('gst', 'mime_types')
-
-        except Exception as e:
-            gst_version = 'GStreamer not available'
-            logger.debug(_('Media support using {} is disabled.').format('GStreamer'))
-            logger.debug(_('Caused by ') + type(e).__name__ + ': ' + str(e))
+        if not self.conf.getboolean('gstreamer', 'enabled'):
+            gst_version = 'GStreamer disabled'
+        else:
+            try:
+                from pympress.media_overlays.gst_backend import GstOverlay
+                gst_version = GstOverlay.setup_backend(self.conf.getlist('gstreamer', 'init_options'))
+            except Exception as e:
+                gst_version = 'GStreamer not available'
+                logger.debug(_('Media support using {} is disabled.').format('GStreamer'))
+                logger.debug(_('Caused by ') + type(e).__name__ + ': ' + str(e))
+            else:
+                self._backends['gstreamer'] = GstOverlay
+                self.types_list['gstreamer'] = self.conf.getlist('gstreamer', 'mime_types')
 
 
-        try:
-            from pympress.media_overlays.vlc_backend import VlcOverlay
-
-            vlc_version = VlcOverlay.setup_backend(self.conf.getlist('vlc', 'init_options'))
-            self._backends['vlc'] = VlcOverlay
-            if self.conf.getboolean('vlc', 'enabled'):
+        if not self.conf.getboolean('vlc', 'enabled'):
+            vlc_version = 'VLC disabled'
+        else:
+            try:
+                from pympress.media_overlays.vlc_backend import VlcOverlay
+                vlc_version = VlcOverlay.setup_backend(self.conf.getlist('vlc', 'init_options'))
+            except Exception as e:
+                vlc_version = 'VLC not available'
+                logger.debug(_('Media support using {} is disabled.').format('VLC'))
+                logger.debug(_('Caused by ') + type(e).__name__ + ': ' + str(e))
+            else:
+                self._backends['vlc'] = VlcOverlay
                 self.types_list['vlc'] = self.conf.getlist('vlc', 'mime_types')
 
-        except Exception as e:
-            vlc_version = 'VLC not available'
-            logger.debug(_('Media support using {} is disabled.').format('VLC'))
-            logger.debug(_('Caused by ') + type(e).__name__ + ': ' + str(e))
-
         self.backend_version = ', '.join([gif_version, gst_version, vlc_version])
         logger.info(_('Media support: ') + self.backend_version)
 
@@ -498,20 +431,17 @@ class Media(object):
     def get_factory(self, mime_type):
         """ Returns a class of type :attr:`~_backend`.
         """
-        if mime_type in {'image/gif', 'image/svg+xml'}:
-            return self._backends['gif']
+        # Search for specific mime type, if no matches fall back to empty lists which mean any mime types
+        options = [backend for backend, mime_types in self.types_list.items() if mime_type in mime_types]
+        if not options:
+            options = [backend for backend, mime_types in self.types_list.items() if len(mime_types) == 0]
 
-        # Search for specific mime type
-        for backend, mime_types in self.types_list.items():
-            if mime_type in mime_types:
-                return self._backends[backend]
+        if not options:
+            return None
 
-        # Search for empty list, meaning fallback
-        for backend, mime_types in self.types_list.items():
-            if len(mime_types) == 0:
-                return self._backends[backend]
-
-        return None
+        # Prefer more stable backends with less external dependencies
+        priority = ['gif', 'vlc', 'gstreamer'] if util.IS_WINDOWS else ['gif', 'gstreamer', 'vlc']
+        return self._backends[sorted(options, key=priority.index)[0]]
 
 
 
@@ -774,6 +704,12 @@ class FileWatcher(object):
     # `int` that is a GLib timeout id to delay the callback
     timeout = 0
 
+    #: The :class:`~pathlib.Path` to the file being watched
+    path = None
+
+    #: Callback to be called on file changes, usually connected to :meth:`~pympress.ui.UI.reload_document`
+    callback = lambda: None
+
     def __init__(self):
         try:
             from watchdog.observers import Observer
@@ -814,17 +750,17 @@ class FileWatcher(object):
         self.stop_watching()
 
         scheme, path = uri.split('://', 1)
-        path = url2pathname(path)
-        directory = os.path.dirname(path)
+        self.path = pathlib.Path(url2pathname(path))
         if scheme != 'file':
             logger.error('Impossible to watch files with {} schemes'.format(scheme), exc_info = True)
             return
 
-        self.monitor.on_modified = lambda e: self._enqueue(callback, *args, **kwargs) if e.src_path == path else None
+        self.callback = lambda: callback(*args, **kwargs)
+        self.monitor.on_modified = self._enqueue
         try:
-            self.observer.schedule(self.monitor, directory)
+            self.observer.schedule(self.monitor, str(self.path.parent), recursive=False)
         except OSError:
-            logger.error('Impossible to open dir at {}'.format(directory), exc_info = True)
+            logger.error('Impossible to open dir at {}'.format(str(self.path.parent)), exc_info = True)
 
 
     def stop_watching(self):
@@ -833,23 +769,22 @@ class FileWatcher(object):
         self.observer.unschedule_all()
 
 
-    def _enqueue(self, callback, *args, **kwargs):
-        """ Do not call callback directly, instead delay as to avoid repeated calls in short periods of time.
+    def _enqueue(self, event):
+        """ Call callback with delay, to avoid repeated calls in short periods of time.
 
         Args:
-            callback (`function`): callback to call with all the further arguments
+            event (:class:`~watchdog.events.FileSystemEvent`): the event that caused the callback to be triggered
         """
+        if event.src_path != str(self.path):
+            return
         if self.timeout:
             GLib.Source.remove(self.timeout)
-        self.timeout = GLib.timeout_add(200, self._call, callback, *args, **kwargs)
+        self.timeout = GLib.timeout_add(200, self._call)
 
 
-    def _call(self, callback, *args, **kwargs):
+    def _call(self):
         """ Call the callback.
-
-        Args:
-            callback (`function`): callback to call with all the further arguments
         """
         if self.timeout:
             self.timeout = 0
-        callback(*args, **kwargs)
+        self.callback()
