@@ -27,38 +27,14 @@
 ------------------------------------------------------------------------------------
 """
 
-from __future__ import print_function, unicode_literals
-
 import logging
 logger = logging.getLogger(__name__)
 
-import ctypes
-
-from pympress import builder
+import gi
+gi.require_version('Gtk', '3.0')
 from gi.repository import GLib
 
-
-def get_window_handle(window):
-    """ Uses ctypes to call gdk_win32_window_get_handle which is not available in python gobject introspection porting.
-
-    Solution from http://stackoverflow.com/a/27236258/1387346
-
-    Args:
-        window (:class:`~Gdk.Window`): The window for which we want to get the handle
-
-    Returns:
-        The handle to the win32 window
-    """
-    # get the c gpointer of the gdk window
-    ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
-    ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
-    drawingarea_gpointer = ctypes.pythonapi.PyCapsule_GetPointer(window.__gpointer__, None)
-    # get the win32 handle
-    gdkdll = ctypes.CDLL('libgdk-3-0.dll')
-    handle_getter = gdkdll.gdk_win32_window_get_handle
-    handle_getter.restype = ctypes.c_void_p
-    handle_getter.argtypes = [ctypes.c_void_p]
-    return handle_getter(drawingarea_gpointer)
+from pympress import builder
 
 
 class VideoOverlay(builder.Builder):
@@ -66,9 +42,9 @@ class VideoOverlay(builder.Builder):
 
     Args:
         container (:class:`~Gtk.Overlay`): The container with the slide, at the top of which we add the movie area
-        show_controls (`bool`): whether to display controls on the video player
         page_type (:class:`~pympress.document.PdfPage`): the part of the page to display
-        relative_margins (:class:`~Poppler.Rectangle`): the margins defining the position of the video in the frame.
+        action_map (:class:`~Gio.ActionMap`): the action map that contains the actions for this media
+        media (:class:`~pympress.document.Media`): the object defining the properties of the video such as position etc.
     """
     #: :class:`~Gtk.Overlay` that is the parent of the VideoOverlay widget.
     parent = None
@@ -86,6 +62,8 @@ class VideoOverlay(builder.Builder):
     relative_margins = None
     #: `bool` that tracks whether we should play automatically
     autoplay = False
+    #: `bool` that tracks whether we should play after we finished playing
+    repeat = False
 
     #: `bool` that tracks whether the user is dragging the position
     dragging_position = False
@@ -99,21 +77,26 @@ class VideoOverlay(builder.Builder):
     #: :class:`~Gio.ActionMap` containing the actios for this video overlay
     action_map = None
 
-    def __init__(self, container, show_controls, relative_margins, page_type, action_map):
+    def __init__(self, container, page_type, action_map, media):
         super(VideoOverlay, self).__init__()
 
         self.parent = container
-        self.relative_page_margins = tuple(getattr(relative_margins, v) for v in ('x1', 'y2', 'x2', 'y1'))
+        self.relative_page_margins = tuple(getattr(media.relative_margins, v) for v in ('x1', 'y2', 'x2', 'y1'))
         self.update_margins_for_page(page_type)
 
         self.load_ui('media_overlay')
-        self.toolbar.set_visible(show_controls)
+        self.toolbar.set_visible(media.show_controls)
 
         self.connect_signals(self)
 
         # medias, here the actions are scoped to the current widget
         self.action_map = action_map
         self.media_overlay.insert_action_group('media', self.action_map)
+        self.set_file(media.filename)
+
+        self.autoplay = media.autoplay
+        self.repeat = media.repeat
+        # TODO: handle poster
 
 
     def handle_embed(self, mapped_widget):
@@ -171,6 +154,15 @@ class VideoOverlay(builder.Builder):
         return self.action_map.lookup_action('pause').activate()
 
 
+    def handle_end(self):
+        """ End of the stream reached: restart if looping, otherwise hide overlay
+        """
+        if not self.repeat:
+            self.action_map.lookup_action('stop').activate()
+        else:
+            self.action_map.lookup_action('set_time').activate(GLib.Variant.new_double(0))
+
+
     def update_margins_for_page(self, page_type):
         """ Recalculate the margins around the media in the event of a page type change.
 
@@ -221,7 +213,7 @@ class VideoOverlay(builder.Builder):
         """ Sets the media file to be played by the widget.
 
         Args:
-            filepath (`str`): The path to the media file path
+            filepath (`pathlib.Path`): The path to the media file path
         """
         raise NotImplementedError
 
