@@ -23,8 +23,6 @@
 ---------------------------------------------------------------------------------------
 """
 
-from __future__ import print_function, unicode_literals
-
 import logging
 logger = logging.getLogger(__name__)
 
@@ -46,27 +44,24 @@ class GstOverlay(base.VideoOverlay):
     #: A :class:`~Gst.Playbin` to be play videos
     playbin = None
 
-    #: `str` holding the path to the current file
-    uri = None
-
     def __init__(self, *args, **kwargs):
-        super(GstOverlay, self).__init__(*args, **kwargs)
-
         # Create GStreamer playbin
         self.playbin = Gst.ElementFactory.make('playbin', None)
         self.sink = Gst.ElementFactory.make('gtksink', None)
         self.playbin.set_property('video-sink', self.sink)
 
+        super(GstOverlay, self).__init__(*args, **kwargs)
+
         self.media_overlay.remove(self.movie_zone)
         self.media_overlay.pack_start(self.sink.props.widget, True, True,  0)
         self.media_overlay.reorder_child(self.sink.props.widget, 0)
-        self.sink.props.widget.show()
+        self.sink.props.widget.hide()
 
         # Create bus to get events from GStreamer playin
         bus = self.playbin.get_bus()
         bus.add_signal_watch()
         bus.enable_sync_message_emission()
-        bus.connect('message::eos', lambda *args: GLib.idle_add(self.do_hide))
+        bus.connect('message::eos', lambda *args: GLib.idle_add(self.handle_end))
         bus.connect('message::error', lambda _, msg: logger.error('{} {}'.format(*msg.parse_error())))
         bus.connect('message::async-done', self.on_play)
         bus.connect('message::duration-changed', lambda *args: GLib.idle_add(self.do_update_duration))
@@ -85,9 +80,10 @@ class GstOverlay(base.VideoOverlay):
         """ Sets the media file to be played by the widget.
 
         Args:
-            filepath (`str`): The path to the media file path
+            filepath (`pathlib.Path`): The path to the media file path
         """
-        self.uri = 'file://' + filepath
+        self.playbin.set_property('uri', filepath.as_uri())
+        self.playbin.set_state(Gst.State.READY)
 
 
     def mute(self, value):
@@ -105,6 +101,7 @@ class GstOverlay(base.VideoOverlay):
         """
         GLib.idle_add(self.do_update_duration)
         GLib.timeout_add(200, self.do_update_time)
+        self.sink.props.widget.show()
 
 
     def do_update_duration(self, *args):
@@ -131,7 +128,6 @@ class GstOverlay(base.VideoOverlay):
         Returns:
             `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
         """
-        self.playbin.set_property('uri', self.uri)
         self.playbin.set_state(Gst.State.PLAYING)
 
         return False
@@ -154,6 +150,8 @@ class GstOverlay(base.VideoOverlay):
         """ Stops playing in the backend player.
         """
         self.playbin.set_state(Gst.State.NULL)
+        self.playbin.set_state(Gst.State.READY)
+        self.sink.props.widget.hide()
 
         return False
 
