@@ -23,13 +23,12 @@
 ---------------------------------------------------------------------
 """
 
-from __future__ import print_function, unicode_literals
-
 import logging
 logger = logging.getLogger(__name__)
 
 import os
 import vlc
+import ctypes
 
 import gi
 gi.require_version('Gtk', '3.0')
@@ -39,6 +38,29 @@ from pympress.util import IS_WINDOWS
 from pympress.media_overlays import base
 
 
+def get_window_handle(window):
+    """ Uses ctypes to call gdk_win32_window_get_handle which is not available in python gobject introspection porting.
+
+    Solution from http://stackoverflow.com/a/27236258/1387346
+
+    Args:
+        window (:class:`~Gdk.Window`): The window for which we want to get the handle
+
+    Returns:
+        The handle to the win32 window
+    """
+    # get the c gpointer of the gdk window
+    ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
+    ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
+    drawingarea_gpointer = ctypes.pythonapi.PyCapsule_GetPointer(window.__gpointer__, None)
+    # get the win32 handle
+    gdkdll = ctypes.CDLL('libgdk-3-0.dll')
+    handle_getter = gdkdll.gdk_win32_window_get_handle
+    handle_getter.restype = ctypes.c_void_p
+    handle_getter.argtypes = [ctypes.c_void_p]
+    return handle_getter(drawingarea_gpointer)
+
+
 class VlcOverlay(base.VideoOverlay):
     """ Simple VLC widget.
 
@@ -52,10 +74,11 @@ class VlcOverlay(base.VideoOverlay):
         self.player = self._instance.media_player_new()  # before loading UI, needed to connect "map" signal
 
         super(VlcOverlay, self).__init__(*args, **kwargs)
+        # Simple black background painting to avoid glitching outside of video area
+        self.movie_zone.connect('draw', self.paint_backdrop)
 
         event_manager = self.player.event_manager()
-        event_manager.event_attach(vlc.EventType.MediaPlayerEndReached,
-                                   lambda e, act=self.action_map.lookup_action('stop'): GLib.idle_add(act.activate))
+        event_manager.event_attach(vlc.EventType.MediaPlayerEndReached, lambda e: GLib.idle_add(self.handle_end))
         event_manager.event_attach(vlc.EventType.MediaPlayerLengthChanged,
                                    lambda e: self.update_range(self.player.get_length() / 1000. or 1.))
         event_manager.event_attach(vlc.EventType.MediaPlayerTimeChanged,
@@ -72,9 +95,10 @@ class VlcOverlay(base.VideoOverlay):
             logger.error('No window in which to embed the VLC player!')
             return False
         elif IS_WINDOWS:
-            self.player.set_hwnd(base.get_window_handle(window))  # get_property('window')
+            self.player.set_hwnd(get_window_handle(window))  # get_property('window')
         else:
             self.player.set_xwindow(window.get_xid())
+        self.movie_zone.queue_draw()
         return False
 
 
@@ -91,9 +115,18 @@ class VlcOverlay(base.VideoOverlay):
         """ Sets the media file to be played by the widget.
 
         Args:
-            filepath (`str`): The path to the media file path
+            filepath (`pathlib.Path`): The path to the media file path
         """
-        self.player.set_media(self._instance.media_new(filepath))
+        self.player.set_media(self._instance.media_new(filepath.resolve().as_uri()))
+
+
+    def handle_end(self):
+        """ End of the stream reached: restart if looping, otherwise hide overlay
+        """
+        self.action_map.lookup_action('stop').activate()
+        if self.repeat:
+            self.action_map.lookup_action('set_time').activate(GLib.Variant.new_double(0))
+            self.action_map.lookup_action('play').activate()
 
 
     def mute(self, value):
@@ -115,9 +148,31 @@ class VlcOverlay(base.VideoOverlay):
             `bool`: `True` iff this function should be run again (:func:`~GLib.idle_add` convention)
         """
         self.player.play()
+        self.movie_zone.queue_draw()
         return False
 
 
+    def paint_backdrop(self, widget, context):
+        """ Draw behind/around the video, aka the black bars
+
+        Args:
+            widget (:class:`~Gtk.Widget`):  the widget to update
+            context (:class:`~cairo.Context`):  the Cairo context (or `None` if called directly)
+        """
+        context.save()
+        context.set_source_rgb(0, 0, 0)
+        context.fill()
+        context.paint()
+        context.restore()
+
+
+    def show(self):
+        """ Bring the widget to the top of the overlays if necessary − also force redraw of movie zone
+        """
+        super(VlcOverlay, self).show()
+        self.movie_zone.queue_draw()
+
+
     def do_play_pause(self):
         """ Toggle pause mode of the media.
 
