@@ -23,8 +23,6 @@
 ---------------------------------------------------------------------
 """
 
-from __future__ import print_function, unicode_literals
-
 import logging
 logger = logging.getLogger(__name__)
 
@@ -79,7 +77,7 @@ class Scribbler(builder.Builder):
     #: The `list` containing the radio buttons :class:`~Gtk.ModelButton`
     scribble_preset_buttons = []
 
-    #: The position of the mouse on the slide as `tuple` of `float`s
+    #: The position of the mouse on the slide as `tuple` of `float`
     mouse_pos = None
     #: A :class:`~cairo.Surface` to hold drawn highlights
     scribble_cache = None
@@ -176,7 +174,7 @@ class Scribbler(builder.Builder):
         masks = [cairo.ImageSurface.create_from_png(util.get_icon_path('marker_fill_{}.png'.format(n))) for n in ms]
 
         self.marker_surfaces = list(zip(icons, masks))
-        self.eraser_surface = cairo.ImageSurface.create_from_png(util.get_icon_path('eraser.png'))
+        self.eraser_surface = cairo.ImageSurface.create_from_png(str(util.get_icon_path('eraser.png')))
 
         # Load color and active pen preferences. Pen 0 is the eraser.
         self.color_width = [(Gdk.RGBA(0, 0, 0, 0), 150)] + list(zip(
@@ -673,11 +671,12 @@ class Scribbler(builder.Builder):
         if not self.scribbling_mode:
             return False
 
+        self.scribbling_mode = False
+
         extras.Cursor.set_cursor(self.scribble_p_da, 'default')
         self.load_layout(None)
         self.off_render.add(self.scribble_overlay)
 
-        self.scribbling_mode = False
         self.get_application().lookup_action('highlight').change_state(GLib.Variant.new_boolean(self.scribbling_mode))
         self.pen_action.set_enabled(self.scribbling_mode)
 
