@@ -31,14 +31,15 @@ current and the next page, as well as a time counter and a clock.
 Both windows are managed by the :class:`~pympress.ui.UI` class.
 """
 
-from __future__ import print_function, unicode_literals
-
 import logging
 logger = logging.getLogger(__name__)
 
-import os.path
+import pathlib
+import math
 import sys
 import gc
+from urllib.request import url2pathname
+from urllib.parse import urlsplit
 
 import gi
 import cairo
@@ -46,7 +47,7 @@ gi.require_version('Gtk', '3.0')
 from gi.repository import GObject, Gtk, Gdk, GLib, GdkPixbuf, Gio
 
 
-from pympress import document, surfacecache, util, pointer, scribble, builder, talk_time, extras, editable_label
+from pympress import document, surfacecache, util, pointer, scribble, builder, talk_time, dialog, extras, editable_label
 
 
 class UI(builder.Builder):
@@ -70,11 +71,15 @@ class UI(builder.Builder):
     #: :class:`~Gtk.DrawingArea` for the current slide in the Presenter window.
     p_da_notes = None
 
-    #: :class:`~Gtk.AspectFrame` for the next slide in the Presenter window.
-    p_frame_next = None
+    #: :class:`~Gtk.Grid` for the next slide(s) in the Presenter window.
+    grid_next = None
+    #: `list` of :class:`~Gtk.AspectFrame` for the next slide in the Presenter window.
+    p_frames_next = None
     #: :class:`~Gtk.DrawingArea` for the next slide in the Presenter window.
-    p_da_next = None
-    #: :class:`~Gtk.AspectFrame` for the current slide copy in the Presenter window.
+    p_das_next = None
+    #: `int` the number of next slides currently on display in the “Next slides” pane, initialized to the maximal number
+    next_frames_count = 16
+    #: `list` of :class:`~Gtk.AspectFrame` for the current slide copy in the Presenter window.
     p_frame_cur = None
     #: :class:`~Gtk.DrawingArea` for the current slide copy in the Presenter window.
     p_da_cur = None
@@ -144,13 +149,17 @@ class UI(builder.Builder):
 
     #: :class:`~pympress.editable_label.EstimatedTalkTime` to set estimated/remaining talk time
     est_time = None
-    #: :class:`~pympress.extras.TimingReport` popup to show how much time was spent on which part
+    #: :class:`~pympress.dialog.TimingReport` popup to show how much time was spent on which part
     timing = None
     #: :class:`~pympress.talk_time.TimeCounter` clock tracking talk time (elapsed, and remaining)
     talk_time = None
 
     #: A :class:`~Gtk.ShortcutsWindow` to show the shortcuts
     shortcuts_window = None
+    #: :class:`~pympress.dialog.LayoutEditor` popup to configure the layouts of the presenter window
+    layout_editor = None
+    #: :class:`~pympress.dialog.AutoPlay` popup to configure automatic playing
+    autoplay = None
 
     #: A :class:`~Gtk.AccelGroup` to store the shortcuts
     accel_group = None
@@ -223,13 +232,19 @@ class UI(builder.Builder):
             'validate-input':        dict(activate=self.validate_current_input),
             'cancel-input':          dict(activate=self.cancel_current_input),
             'align-content':         dict(activate=self.adjust_frame_position),
+            'next-frames':           dict(activate=self.reconfigure_next_frames, parameter_type=int,
+                                          state=self.next_frames_count),
         })
 
         self.setup_actions({
             'open-file':         dict(activate=self.open_file, parameter_type=str),
             'close-file':        dict(activate=self.close_file),
+            'save-file':         dict(activate=self.save_file),
+            'save-file-as':      dict(activate=self.save_file_as),
             'pick-file':         dict(activate=self.pick_file),
             'list-recent-files': dict(change_state=self.populate_recent_menu, state=False),
+            'page':              dict(activate=lambda gaction, param: self.goto_page(param.get_int64()),
+                                      parameter_type=int, state=self.current_page),
             'next-page':         dict(activate=self.doc_goto_next),
             'next-label':        dict(activate=self.doc_label_next),
             'prev-page':         dict(activate=self.doc_goto_prev),
@@ -247,8 +262,10 @@ class UI(builder.Builder):
         self.laser = pointer.Pointer(self.config, self)
         self.est_time = editable_label.EstimatedTalkTime(self)
         self.page_number = editable_label.PageNumber(self, self.config.getboolean('presenter', 'scroll_number'))
-        self.timing = extras.TimingReport(self)
-        self.talk_time = talk_time.TimeCounter(self, self.est_time, self.timing)
+        self.timing = dialog.TimingReport(self)
+        self.autoplay = dialog.AutoPlay(self)
+        self.talk_time = talk_time.TimeCounter(self, self.est_time, self.timing, self.autoplay)
+        self.layout_editor = dialog.LayoutEditor(self, self.config)
         self.file_watcher = extras.FileWatcher()
         self.config.register_actions(self)
 
@@ -335,10 +352,15 @@ class UI(builder.Builder):
         pane_handles = self.replace_layout(layout, self.p_central, self.placeable_widgets, self.on_pane_event)
         self.pane_handle_pos.update(pane_handles)
 
+        self.p_frames_next = [self.get_object('p_frame_next{}'.format(n)) for n in range(self.next_frames_count)]
+        self.p_das_next = [self.get_object('p_da_next{}'.format(n)) for n in range(self.next_frames_count)]
+        self.reconfigure_next_frames(None, GLib.Variant.new_int64(self.config.getint('presenter', 'next_slide_count')))
+
         slide_type = self.notes_mode.complement()
         self.cache.add_widget(self.p_da_cur, slide_type)
         self.cache.add_widget(self.p_da_cur, slide_type, zoomed = True)
-        self.cache.add_widget(self.p_da_next, slide_type)
+        # A single cache for all next slides
+        self.cache.add_widget(self.p_das_next[0], slide_type)
         self.cache.add_widget(self.p_da_notes, self.notes_mode, prerender_enabled = bool(self.notes_mode))
         self.cache.add_widget(self.scribbler.scribble_p_da, slide_type, prerender_enabled = False)
         self.cache.add_widget(self.scribbler.scribble_p_da, slide_type, zoomed = True)
@@ -354,6 +376,37 @@ class UI(builder.Builder):
         self.p_win.insert_action_group('win', None)
 
 
+    def reconfigure_next_frames(self, gaction, param):
+        """ Set the number of next frames to preview the the “next slides” panel
+
+        Args:
+            gaction (:class:`~Gio.Action`): the action triggering the call
+            param (:class:`~GLib.Variant`): the number of slides as a GVariant
+        """
+        n_frames = param.get_int64()
+        if n_frames < 1:
+            n_frames = 1
+        if n_frames > 16:
+            n_frames = 16
+
+        for n in range(self.next_frames_count):
+            self.grid_next.remove(self.p_frames_next[n])
+
+        for n in range(4):
+            self.grid_next.remove_row(n)
+            self.grid_next.remove_column(n)
+
+        rows = int(round(math.sqrt(n_frames)))
+        cols = (n_frames + rows - 1) // rows
+
+        for n in range(n_frames):
+            self.grid_next.attach(self.p_frames_next[n], n % cols, n // cols, 1, 1)
+
+        self.next_frames_count = n_frames
+        self.app.set_action_state('next-frames', n_frames)
+        self.config.set('presenter', 'next_slide_count', str(n_frames))
+
+
     def setup_screens(self):
         """ Sets up the position of the windows.
         """
@@ -478,7 +531,7 @@ class UI(builder.Builder):
         if not event.send_event:
             return
 
-        self.cache.resize_widget(widget.get_name(), event.width, event.height)
+        self.cache.resize_widget(widget.get_name().rstrip('0123456789'), event.width, event.height)
 
         if widget is self.c_da:
             self.medias.resize('content')
@@ -526,14 +579,15 @@ class UI(builder.Builder):
         """
         self.resize_panes = False
         self.p_da_cur.queue_draw()
-        self.p_da_next.queue_draw()
+        for da in self.p_das_next:
+            da.queue_draw()
         if self.notes_mode:
             self.p_da_notes.queue_draw()
         if self.redraw_timeout:
             self.redraw_timeout = 0
 
-        self.config.update_layout('highlight' if self.scribbler.scribbling_mode else self.layout_name(self.notes_mode),
-                                  self.p_central.get_children()[0], self.pane_handle_pos)
+        self.config.update_layout_from_widgets(self.layout_name(self.notes_mode), self.p_central.get_children()[0],
+                                               self.pane_handle_pos)
 
 
     def on_pane_event(self, widget, evt):
@@ -584,9 +638,9 @@ class UI(builder.Builder):
         about.set_copyright(_('Contributors:') + '\n' + pympress['contributors'])
         about.set_comments(_('pympress is a little PDF reader written in Python ' +
                              'using Poppler for PDF rendering and GTK for the GUI.\n') +
-                           _('Some preferences are saved in ') + self.config.path_to_config() + '\n' +
-                           _('Resources are loaded from ') + os.path.dirname(util.get_locale_dir()) + '\n' +
-                           _('The log is written to ') + util.get_log_path() + '\n\n' +
+                           _('Some preferences are saved in ') + str(self.config.path_to_config()) + '\n' +
+                           _('Resources are loaded from ') + str(util.get_locale_dir().parent) + '\n' +
+                           _('The log is written to ') + str(util.get_log_path()) + '\n\n' +
                            _('Media support uses {}.').format(self.medias.backend_version) + '\n' +
                            _('Python version {}').format(sys.version))
         about.set_website('https://github.com/Cimbali/pympress')
@@ -613,6 +667,13 @@ class UI(builder.Builder):
             page (`int`): the page at which to start the presentation
             reloading (`bool`): whether we are reloading or detecting stuff from the document
         """
+        if self.unsaved_changes(reloading):
+            # If we keep the unsaved changes and were prompted by automatic file change,
+            # stop watching for file changes in the future.
+            if reloading:
+                self.file_watcher.stop_watching()
+            return
+
         run_gc = self.doc.doc is not None
         try:
             self.doc = document.Document.create(self, doc_uri)
@@ -654,6 +715,7 @@ class UI(builder.Builder):
         self.cache.swap_document(self.doc)
         self.page_number.set_last(self.doc.pages_number())
         self.page_number.enable_labels(self.doc.has_labels())
+        self.autoplay.set_doc_pages(self.doc.pages_number())
         self.medias.purge_media_overlays()
         self.timing.set_document_metadata(self.doc.get_structure().copy(), self.doc.page_labels[:])
 
@@ -716,12 +778,88 @@ class UI(builder.Builder):
             info (`int`): info on the target
             time (`int`): time of the drop
         """
-        received = data.get_text()
-        if received.startswith('file://'):
-            received = received[len('file://'):]
+        uri_parts = urlsplit(data.get_text())
+        path = pathlib.Path(url2pathname(uri_parts.path) if uri_parts.scheme else uri_parts.path)
 
-        if os.path.isfile(received) and received.lower().endswith('.pdf'):
-            self.swap_document(os.path.abspath(received))
+        if path.is_file() and path.suffix.lower() == '.pdf':
+            self.swap_document(path.as_uri())
+
+
+    def unsaved_changes(self, reload=False):
+        """ Prompt the user about what to do with changes in the document: save, discard, or cancel action
+
+        Args:
+            reload (`bool`): The unsaved changes is prompted by reloading the file
+
+        Returns:
+            `bool`: `True` iff we need to cancel the current action
+        """
+        if not self.doc.has_changes():
+            return False
+
+        dialog = Gtk.MessageDialog(title=_('Unsaved changes'), transient_for=self.p_win, type=Gtk.MessageType.WARNING)
+        dialog.set_position(Gtk.WindowPosition.CENTER)
+        if not reload:
+            dialog.set_markup(_('Save changes before closing?'))
+            dialog.format_secondary_markup(_('Unsaved changes will be lost'))
+
+            dialog.add_button(Gtk.STOCK_SAVE, Gtk.ResponseType.YES)
+            dialog.add_button(_('_Discard'), Gtk.ResponseType.NO).set_image(
+                Gtk.Button.new_from_icon_name('edit-delete', Gtk.IconSize.BUTTON).get_image()
+            )
+            dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
+            dialog.set_default_response(Gtk.ResponseType.YES)
+        else:
+            dialog.set_markup(_('The open file was modified outside of pympress but you have made unsaved changes.') +
+                              '\n<b>' + _('Overwrite changes instead of reloading?') + '</b>')
+            dialog.format_secondary_markup(_('Saving changes will overwrite the changed file!') + '\n' +
+                                           _('Unsaved changes will be lost.'))
+
+            dialog.add_button(_('Overwrite'), Gtk.ResponseType.YES).set_image(
+                Gtk.Button.new_from_icon_name('document-save', Gtk.IconSize.BUTTON).get_image()
+            )
+            dialog.add_button(_('Reload'), Gtk.ResponseType.NO).set_image(
+                Gtk.Button.new_from_icon_name('view-refresh', Gtk.IconSize.BUTTON).get_image()
+            )
+            dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
+            dialog.set_default_response(Gtk.ResponseType.YES)
+
+        response = dialog.run()
+        dialog.destroy()
+
+        if response == Gtk.ResponseType.YES:
+            self.doc.save_changes()
+
+        return response == Gtk.ResponseType.CANCEL or (reload and Gtk.ResponseType == Gtk.ResponseType.YES)
+
+
+    def save_file_as(self, *args):
+        """ Remove the current document.
+        """
+        # Use a GTK file dialog to choose file
+        dialog = Gtk.FileChooserDialog(title = _('Save as...'), transient_for = self.p_win,
+                                       action = Gtk.FileChooserAction.SAVE)
+        dialog.add_buttons(Gtk.STOCK_SAVE_AS, Gtk.ResponseType.OK)
+        dialog.set_default_response(Gtk.ResponseType.OK)
+        dialog.set_position(Gtk.WindowPosition.CENTER)
+
+        filter = Gtk.FileFilter()
+        filter.set_name(_('PDF files'))
+        filter.add_mime_type('application/pdf')
+        filter.add_pattern('*.pdf')
+        dialog.add_filter(filter)
+
+        filter = Gtk.FileFilter()
+        filter.set_name(_('All files'))
+        filter.add_pattern('*')
+        dialog.add_filter(filter)
+
+        response = dialog.run()
+
+        if response == Gtk.ResponseType.OK:
+            self.doc.save_changes(dialog.get_uri())
+
+        dialog.destroy()
 
 
     def pick_file(self, *args):
@@ -753,17 +891,22 @@ class UI(builder.Builder):
         dialog.destroy()
 
 
-    def error_opening_file(self, filename):
+    def error_opening_file(self, uri):
         """ Remove the current document.
+
+        Args:
+            uri (`str`): the URI of the document
         """
         # Check if the path is valid
-        if not os.path.exists(filename):
+        uri_parts = urlsplit(uri)
+        filename = pathlib.Path(url2pathname(uri_parts.path))
+        if uri_parts.scheme == 'file' and not filename.exists():
             msg = _('Could not find the file "{}"').format(filename)
         else:
-            msg = _('Error opening the file "{}"').format(filename)
-        dialog = Gtk.MessageDialog(transient_for = self.p_win, flags = Gtk.DialogFlags.MODAL,
-                                   message_type = Gtk.MessageType.ERROR, message_format = msg)
-        dialog.add_buttons(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
+            msg = _('Error opening the file "{}"').format(uri)
+        dialog = Gtk.MessageDialog(transient_for=self.p_win, flags=Gtk.DialogFlags.MODAL,
+                                   message_type=Gtk.MessageType.ERROR, message_format=msg)
+        dialog.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
         dialog.set_position(Gtk.WindowPosition.CENTER)
         dialog.run()
         dialog.destroy()
@@ -775,10 +918,16 @@ class UI(builder.Builder):
         self.swap_document(None)
 
 
+    def save_file(self, *args):
+        """ Remove the current document.
+        """
+        self.doc.save_changes()
+
+
     def open_file(self, gaction, target):
         """ Open a document.
 
-        Returns:
+        Args:
             gaction (:class:`~Gio.Action`): the action triggering the call
             target (:class:`~GLib.Variant`): the file to open as a string variant
         """
@@ -788,8 +937,8 @@ class UI(builder.Builder):
     def get_notes_mode(self):
         """ Simple getter.
 
-        Returns (:class:`~pympress.document.PdfPage`):
-            Truthy when we split slides in content + notes
+        Returns:
+             :class:`~pympress.document.PdfPage`: Truthy when we split slides in content + notes
         """
         return self.notes_mode
 
@@ -814,18 +963,19 @@ class UI(builder.Builder):
             self.goto_page(display_page_num - 1)
 
 
-    def goto_page(self, page):
+    def goto_page(self, page, autoplay=False):
         """ Handle going to the page passed as argument
 
         Args:
             page (`int`): the page to which to go. Will be clipped to document pages.
+            autoplay (`bool`): whether this page change was triggered automatically
         """
         self.preview_page = self.doc.goto(page)
 
         if not self.page_number.editing:
             self.current_page = self.preview_page
 
-        self.do_page_change()
+        self.do_page_change(autoplay=autoplay)
 
 
     def doc_goto_prev(self, gaction=None, param=None):
@@ -915,7 +1065,7 @@ class UI(builder.Builder):
         self.goto_page(self.doc.pages_number())
 
 
-    def do_page_change(self, unpause=True):
+    def do_page_change(self, unpause=True, autoplay=False):
         """ Switch to another page and display it.
 
         This is a kind of event which is supposed to be called only from the
@@ -924,17 +1074,19 @@ class UI(builder.Builder):
         Args:
             is_preview (`bool`):  `True` if the page change should not update the content
             unpause (`bool`):  `True` if the page change should unpause the timer, `False` otherwise
+            autoplay (`bool`):  `True` if the page change is triggered automatically, otherwise cancel auto play
         """
         is_preview = self.page_number.editing
         if not is_preview:
             self.preview_page = self.current_page
+            self.app.set_action_state('page', self.current_page + 1)
 
         draw_notes = self.notes_mode
         draw_page = draw_notes.complement()
 
         page_content = self.doc.page(self.current_page)
         page_preview = self.doc.page(self.preview_page)
-        page_next = self.doc.page(self.preview_page + 1)
+        pages_next = [self.doc.page(self.preview_page + n + 1) for n in range(self.next_frames_count)]
         page_notes = self.doc.notes_page(self.preview_page)
 
         # Aspect ratios and queue redraws
@@ -956,19 +1108,20 @@ class UI(builder.Builder):
             self.scribbler.scribble_p_frame.set_property('ratio', content_pr)
             self.scribbler.scribble_p_frame.queue_draw()
 
-        if page_next is not None:
-            next_pr = page_next.get_aspect_ratio(draw_notes)
-            self.p_frame_next.set_property('ratio', next_pr)
+        next_pr = preview_pr  # A default page ratio
+        for page, frame, da in zip(pages_next, self.p_frames_next, self.p_das_next):
+            if page is not None:
+                next_pr = page.get_aspect_ratio(draw_page)
+            frame.set_property('ratio', next_pr)
+            da.queue_draw()
 
-        self.p_da_next.queue_draw()
-
-        self.annotations.add_annotations(page_preview.get_annotations())
+        self.annotations.load_annotations(page_preview)
 
         # Update display -- needs to be different ?
         self.page_number.update_page_numbers(self.preview_page, page_preview.label())
 
         # Prerender the 4 next pages and the 2 previous ones
-        page_max = min(self.doc.pages_number(), self.preview_page + 5)
+        page_max = min(self.doc.pages_number(), self.preview_page + self.next_frames_count + 4)
         page_min = max(0, self.preview_page - 2)
         for p in list(range(self.preview_page + 1, page_max)) + list(range(self.preview_page, page_min, -1)):
             self.cache.prerender(p)
@@ -987,6 +1140,8 @@ class UI(builder.Builder):
         # Start counter if needed
         if unpause:
             self.talk_time.unpause()
+        if not autoplay:
+            self.autoplay.stop_looping()
 
         self.timing.transition(self.preview_page, self.talk_time.current_time())
 
@@ -1014,16 +1169,20 @@ class UI(builder.Builder):
         elif widget is self.p_da_notes:
             # Notes page, aligned with preview
             page = self.doc.notes_page(self.preview_page)
-        else:
-            page = self.doc.page(self.preview_page + 1)
+        elif widget in self.p_das_next:
+            offset = 1 + int(widget.get_name()[len('p_da_next'):])
+            page = self.doc.page(self.preview_page + offset)
             # No next page: just return so we won't draw anything
             if page is None:
                 return
+        else:
+            logger.warning(_('Unknown widget "{}" to draw'.format(widget.getname())))
+            return
 
         if not page.can_render():
             return
 
-        name = widget.get_name()
+        name = widget.get_name().rstrip('0123456789')
         nb = page.number()
         wtype = self.cache.get_widget_type(name)
         ww, wh = widget.get_allocated_width(), widget.get_allocated_height()
@@ -1039,7 +1198,7 @@ class UI(builder.Builder):
 
         pb = self.cache.get(name, nb)
         if pb is None:
-            if self.resize_panes and widget in [self.p_da_next, self.p_da_cur, self.p_da_notes]:
+            if self.resize_panes and widget in self.p_das_next + [self.p_da_cur, self.p_da_notes]:
                 # too slow to render here when resize_panes things
                 return
 
@@ -1130,6 +1289,19 @@ class UI(builder.Builder):
         return False
 
 
+    def on_key_input(self, widget, event):
+        """ Handle key strokes at top level, only for when editing needs to bypass action accelerators
+
+        Args:
+            widget (:class:`~Gtk.Widget`):  the widget which has received the key stroke
+            event (:class:`~Gdk.Event`):  the GTK event, which contains the key stroke details
+        """
+        if self.annotations.key_event(widget, event):
+            return True
+
+        return False
+
+
     def cancel_current_input(self, gaction, param=None):
         """ Handle the action cancelling the input, if applicable.
 
@@ -1145,6 +1317,10 @@ class UI(builder.Builder):
             return True
         elif self.scribbler.try_cancel():
             return True
+        elif self.annotations.try_cancel():
+            return True
+        elif self.autoplay.stop_looping():
+            return True
 
         return False
 
@@ -1165,8 +1341,6 @@ class UI(builder.Builder):
         # send to spinner if it is active
         elif self.page_number.on_scroll(widget, event):
             return True
-        elif self.annotations.on_scroll(widget, event):
-            return True
         else:
             return False
 
@@ -1231,8 +1405,9 @@ class UI(builder.Builder):
             return False
 
         # Where did the event occur?
-        if widget is self.p_da_next:
-            page = self.doc.page(self.preview_page + 1)
+        if widget in self.p_das_next:
+            offset = 1 + int(widget.get_name()[len('p_da_next'):])
+            page = self.doc.page(self.preview_page + offset)
         elif widget is self.p_da_notes:
             page = self.doc.notes_page(self.preview_page)
         elif widget is self.p_da_cur:
@@ -1269,8 +1444,9 @@ class UI(builder.Builder):
             return False
 
         # Where did the event occur?
-        if widget is self.p_da_next:
-            page = self.doc.page(self.preview_page + 1)
+        if widget in self.p_das_next:
+            offset = 1 + int(widget.get_name()[len('p_da_next'):])
+            page = self.doc.page(self.preview_page + offset)
         elif widget is self.p_da_notes:
             page = self.doc.notes_page(self.preview_page)
         elif widget is self.p_da_cur:
@@ -1447,7 +1623,9 @@ class UI(builder.Builder):
         Returns:
             `str`: a string representing the appropriate layout
         """
-        if notes_mode.direction() == 'page number':
+        if self.scribbler.scribbling_mode:
+            return 'highlight'
+        elif notes_mode.direction() == 'page number':
             return 'note_pages'
         elif notes_mode:
             return 'notes'
@@ -1463,9 +1641,12 @@ class UI(builder.Builder):
         """
         if new is None:
             new = self.layout_name(self.notes_mode)
+        else:
+            self.layout_editor.set_current_layout(new)
 
         pane_handles = self.replace_layout(self.config.get_layout(new), self.p_central,
                                            self.placeable_widgets, self.on_pane_event)
+        self.scribbler.adjust_tools_orientation()
         self.pane_handle_pos.update(pane_handles)
 
         # queue visibility of all newly added widgets, make sure visibility is right
@@ -1570,12 +1751,12 @@ class UI(builder.Builder):
             parent = self.p_frame_annot.get_parent()
             if issubclass(type(parent), Gtk.Paned):
                 if parent.get_orientation() == Gtk.Orientation.HORIZONTAL:
-                    size = parent.get_parent().get_allocated_width()
+                    size = parent.get_allocated_width()
                 else:
-                    size = parent.get_parent().get_allocated_height()
+                    size = parent.get_allocated_height()
                 parent.set_position(self.pane_handle_pos[parent] * size)
 
-        self.annotations.add_annotations(self.doc.page(self.preview_page).get_annotations())
+        self.annotations.load_annotations(self.doc.page(self.preview_page))
         gaction.change_state(GLib.Variant.new_boolean(self.show_annotations))
 
         return True
